# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from typing import Any, List

from airbyte_cdk.models import (
    AirbyteStateBlob,
    AirbyteStateMessage,
    AirbyteStateType,
    AirbyteStreamState,
    StreamDescriptor,
)


class StateBuilder:
    def __init__(self) -> None:
        self._state: List[AirbyteStateMessage] = []

    def with_stream_state(self, stream_name: str, state: Any) -> "StateBuilder":
        self._state.append(
            AirbyteStateMessage(
                type=AirbyteStateType.STREAM,
                stream=AirbyteStreamState(
                    stream_state=state
                    if isinstance(state, AirbyteStateBlob)
                    else AirbyteStateBlob(state),
                    stream_descriptor=StreamDescriptor(**{"name": stream_name}),
                ),
            )
        )
        return self

    def build(self) -> List[AirbyteStateMessage]:
        return self._state



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from typing import Any, Iterable, Mapping, Optional

from typing_extensions import deprecated

from airbyte_cdk.sources.streams.core import StreamData
from airbyte_cdk.sources.types import StreamSlice, StreamState


class Retriever:
    """
    Responsible for fetching a stream's records from an HTTP API source.
    """

    @abstractmethod
    def read_records(
        self,
        records_schema: Mapping[str, Any],
        stream_slice: Optional[StreamSlice] = None,
    ) -> Iterable[StreamData]:
        """
        Fetch a stream's records from an HTTP API source

        :param records_schema: json schema to describe record
        :param stream_slice: The stream slice to read data for
        :return: The records read from the API source
        """

    @deprecated("Stream slicing is being moved to the stream level.")
    def stream_slices(self) -> Iterable[Optional[StreamSlice]]:
        """Does nothing as this method is deprecated, so underlying Retriever implementations
        do not need to implement this.
        """
        yield from []

    @property
    @deprecated("State management is being moved to the stream level.")
    def state(self) -> StreamState:
        """
        Does nothing as this method is deprecated, so underlying Retriever implementations
        do not need to implement this.
        """
        return {}

    @state.setter
    @deprecated("State management is being moved to the stream level.")
    def state(self, value: StreamState) -> None:
        """
        Does nothing as this method is deprecated, so underlying Retriever implementations
        do not need to implement this.
        """
        pass



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from typing import Any, Iterable, Mapping, Optional, cast

from airbyte_cdk.sources.declarative.retrievers import Retriever
from airbyte_cdk.sources.declarative.schema import SchemaLoader
from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer_test_read_decorator import (
    StreamSlicerTestReadDecorator,
)
from airbyte_cdk.sources.message import MessageRepository
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.partition_generator import PartitionGenerator
from airbyte_cdk.sources.streams.concurrent.partitions.stream_slicer import StreamSlicer
from airbyte_cdk.sources.types import Record, StreamSlice
from airbyte_cdk.utils.slice_hasher import SliceHasher


# For Connector Builder test read operations, we track the total number of records
# read for the stream so that we can stop reading early if we exceed the record limit.
class RecordCounter:
    def __init__(self) -> None:
        self.total_record_counter = 0

    def increment(self) -> None:
        self.total_record_counter += 1

    def reset(self) -> None:
        self.total_record_counter = 0

    def get_total_records(self) -> int:
        return self.total_record_counter


class DeclarativePartitionFactory:
    def __init__(
        self,
        stream_name: str,
        schema_loader: SchemaLoader,
        retriever: Retriever,
        message_repository: MessageRepository,
        max_records_limit: Optional[int] = None,
    ) -> None:
        """
        The DeclarativePartitionFactory takes a retriever_factory and not a retriever directly. The reason is that our components are not
        thread safe and classes like `DefaultPaginator` may not work because multiple threads can access and modify a shared field across each other.
        In order to avoid these problems, we will create one retriever per thread which should make the processing thread-safe.
        """
        self._stream_name = stream_name
        self._schema_loader = schema_loader
        self._retriever = retriever
        self._message_repository = message_repository
        self._max_records_limit = max_records_limit
        self._record_counter = RecordCounter()

    def create(self, stream_slice: StreamSlice) -> Partition:
        return DeclarativePartition(
            stream_name=self._stream_name,
            schema_loader=self._schema_loader,
            retriever=self._retriever,
            message_repository=self._message_repository,
            max_records_limit=self._max_records_limit,
            stream_slice=stream_slice,
            record_counter=self._record_counter,
        )


class DeclarativePartition(Partition):
    def __init__(
        self,
        stream_name: str,
        schema_loader: SchemaLoader,
        retriever: Retriever,
        message_repository: MessageRepository,
        max_records_limit: Optional[int],
        stream_slice: StreamSlice,
        record_counter: RecordCounter,
    ):
        self._stream_name = stream_name
        self._schema_loader = schema_loader
        self._retriever = retriever
        self._message_repository = message_repository
        self._max_records_limit = max_records_limit
        self._stream_slice = stream_slice
        self._hash = SliceHasher.hash(self._stream_name, self._stream_slice)
        self._record_counter = record_counter

    def read(self) -> Iterable[Record]:
        if self._max_records_limit is not None:
            if self._record_counter.get_total_records() >= self._max_records_limit:
                return
        for stream_data in self._retriever.read_records(
            self._schema_loader.get_json_schema(), self._stream_slice
        ):
            if self._max_records_limit is not None:
                if self._record_counter.get_total_records() >= self._max_records_limit:
                    break

            if isinstance(stream_data, Mapping):
                record = (
                    stream_data
                    if isinstance(stream_data, Record)
                    else Record(
                        data=stream_data,
                        stream_name=self.stream_name(),
                        associated_slice=self._stream_slice,
                    )
                )
                yield record
            else:
                self._message_repository.emit_message(stream_data)

            if self._max_records_limit is not None:
                self._record_counter.increment()

    def to_slice(self) -> Optional[Mapping[str, Any]]:
        return self._stream_slice

    def stream_name(self) -> str:
        return self._stream_name

    def __hash__(self) -> int:
        return self._hash


class StreamSlicerPartitionGenerator(PartitionGenerator):
    def __init__(
        self,
        partition_factory: DeclarativePartitionFactory,
        stream_slicer: StreamSlicer,
        slice_limit: Optional[int] = None,
        max_records_limit: Optional[int] = None,
    ) -> None:
        self._partition_factory = partition_factory

        if slice_limit:
            self._stream_slicer = cast(
                StreamSlicer,
                StreamSlicerTestReadDecorator(
                    wrapped_slicer=stream_slicer,
                    maximum_number_of_slices=slice_limit,
                ),
            )
        else:
            self._stream_slicer = stream_slicer

    def generate(self) -> Iterable[Partition]:
        for stream_slice in self._stream_slicer.stream_slices():
            yield self._partition_factory.create(stream_slice)



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

from dataclasses import InitVar, dataclass
from typing import Any, Iterable, List, Mapping, Optional

import dpath

from airbyte_cdk.sources.declarative.interpolation import InterpolatedString
from airbyte_cdk.sources.declarative.retrievers import Retriever
from airbyte_cdk.sources.types import Config, StreamSlice


@dataclass
class PropertiesFromEndpoint:
    """
    Component that defines the behavior around how to dynamically retrieve a set of request properties from an
    API endpoint. The set retrieved can then be injected into the requests to extract records from an API source.
    """

    property_field_path: List[str]
    retriever: Retriever
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    _cached_properties: Optional[List[str]] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self._property_field_path = [
            InterpolatedString(string=property_field, parameters=parameters)
            for property_field in self.property_field_path
        ]

    def get_properties_from_endpoint(self) -> List[str]:
        if self._cached_properties is None:
            self._cached_properties = list(
                map(
                    self._get_property,  # type: ignore # SimpleRetriever and AsyncRetriever only returns Record. Should we change the return type of Retriever.read_records?
                    self.retriever.read_records(records_schema={}, stream_slice=None),
                )
            )
        return self._cached_properties

    def _get_property(self, property_obj: Mapping[str, Any]) -> str:
        path = [
            node.eval(self.config) if not isinstance(node, str) else node
            for node in self._property_field_path
        ]
        return str(dpath.get(property_obj, path, default=[]))  # type: ignore # extracted will be a MutableMapping, given input data structure



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from typing import Any, List, Mapping, Optional

from airbyte_cdk import AbstractSource
from airbyte_cdk.models import AirbyteStateMessage, ConfiguredAirbyteCatalog, SyncMode
from airbyte_cdk.test.catalog_builder import CatalogBuilder
from airbyte_cdk.test.entrypoint_wrapper import EntrypointOutput, read
from airbyte_cdk.test.models.outcome import ExpectedOutcome


def catalog(stream_name: str, sync_mode: SyncMode) -> ConfiguredAirbyteCatalog:
    """Create a catalog with a single stream."""
    return CatalogBuilder().with_stream(stream_name, sync_mode).build()


def read_records(
    source: AbstractSource,
    config: Mapping[str, Any],
    stream_name: str,
    sync_mode: SyncMode,
    state: Optional[List[AirbyteStateMessage]] = None,
    expecting_exception: bool | None = None,  # Deprecated, use expected_outcome instead.
    *,
    expected_outcome: ExpectedOutcome | None = None,
) -> EntrypointOutput:
    """Read records from a stream."""
    _catalog = catalog(stream_name, sync_mode)
    return read(
        source,
        config,
        _catalog,
        state,
        expecting_exception=expecting_exception,  # Deprecated, for backward compatibility.
        expected_outcome=expected_outcome,
    )



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from typing import Any, MutableMapping


class ConfigTransformation(ABC):
    """
    Implementations of this class define transformations that can be applied to source configurations.
    """

    @abstractmethod
    def transform(
        self,
        config: MutableMapping[str, Any],
    ) -> None:
        """
        Transform a configuration by adding, deleting, or mutating fields directly from the config reference passed in argument.

        :param config: The user-provided configuration to be transformed
        """
